import logging
from fitxf.utils import Logging, Singleton


class LoggingSingleton:

    LOGGER_TYPE_ROTATING_FILE = 'rotating_file'
    LOGGER_TYPE_STD = 'std'

    def __init__(
            self,
            logger_type: str = LOGGER_TYPE_STD,
            logger_name: str = Logging.LOGGER_NAME_DEFAULT,
            filepath: str = None,
            log_level = logging.INFO,
            log_format: str = Logging.LOG_FORMAT_DEFAULT,
            file_max_bytes: int = (1048576 * 100),
            file_backup_count: int = 7,
            # do not propagate to other loggers (prevent duplicate logging) if False
            propagate = False,
            stream = 'stdout',
    ):
        assert logger_type in (self.LOGGER_TYPE_STD, self.LOGGER_TYPE_ROTATING_FILE,)
        if logger_type == self.LOGGER_TYPE_ROTATING_FILE:
            self.logger = Logging.get_logger_rotating_file(
                logger_name = logger_name,
                log_level = log_level,
                filename = filepath,
                log_format = log_format,
                max_bytes = file_max_bytes,
                backup_count = file_backup_count,
                propagate = propagate,
            )
        else:
            self.logger = Logging.get_default_logger(
                log_level = log_level,
                propagate = propagate,
                stream = stream,
            )
        return

    def info(self, msg):
        return self.logger.info(msg)

    def debug(self, msg):
        return self.logger.debug(msg)

    def warning(self, msg):
        return self.logger.warning(msg)

    def warn(self, msg):
        return self.logger.warn(msg)

    def error(self, msg):
        return self.logger.error(msg)

    def critical(self, msg):
        return self.logger.critical(msg)

    @staticmethod
    def get_singleton_logger(
            logger_type: str = LOGGER_TYPE_STD,
            logger_name: str = Logging.LOGGER_NAME_DEFAULT,
            filepath: str = None,
            log_level = logging.INFO,
            log_format: str = Logging.LOG_FORMAT_DEFAULT,
            file_max_bytes: int = (1048576 * 100),
            file_backup_count: int = 7,
            # do not propagate to other loggers (prevent duplicate logging) if False
            propagate = False,
            stream = 'stdout',
    ):
        key_id = 'logger_type=' + str(logger_type) + ';filepath=' + str(filepath) + ';log_level=' + str(log_level) \
                 + ';propagate=' + str(propagate) + ';stream=' + str(stream)
        sgt = Singleton(
            class_type = LoggingSingleton,
        ).get_singleton(
            key_id,
            logger_type,
            logger_name,
            filepath,
            log_level,
            log_format,
            file_max_bytes,
            file_backup_count,
            propagate,
            stream,
        )
        return sgt


if __name__ == '__main__':
    lgr = LoggingSingleton.get_singleton_logger()
    lgr.info('Ok')
    exit(0)



import logging
import numpy as np
import matplotlib.pyplot as plt
from fitxf.utils import Logging
from fitxf.math.utils.LoggingSingleton import LoggingSingleton


class Mdct:

    def __init__(self, logger: Logging = None):
        self.logger = logger if logger is not None else logging.getLogger()
        return

    def getSineSignal(
            self,
            amps,
            freqs,
            t: np.ndarray,
            show_plot: bool = False,
    ):
        x = np.zeros_like(t)
        for a, f in zip(amps, freqs):
            x += a * np.sin(2 * np.pi * f * t)

        if show_plot:
            plt.figure(figsize=(8, 6))
            plt.plot(t, x, 'r')
            plt.xlabel('t')
            plt.ylabel('x')
            plt.grid(True)
            plt.show()

        return x

    def MDCT(
            self,
            # time domain signal
            x: np.ndarray,
            lib: str = 'dot'
    ):
        assert len(x) % 2 == 0
        N = int(len(x) / 2)
        # n is indexed at the columns
        n = np.arange(2 * N)
        # print(n)
        # reshape so that k is indexed at the rows
        k = np.reshape(np.arange(N), shape=(N, 1))
        # print(k)
        A = 0.5 + N/2
        B = 0.5
        # now we have a k x n matrix
        m = np.cos((np.pi / N) * (n + A) * (k + B))
        # x_square = np.array([x.tolist() for i in range(N)])
        # [print(line) for line in np.round(m,3).tolist()]
        y = np.dot(m, x)
        return y

    def IMDCT(
            self,
            y: np.ndarray,
    ):
        N = len(y)
        # k is indexed at the columns
        k = np.arange(N)
        # print(n)
        # reshape so that n is indexed at the rows
        n = np.reshape(np.arange(2 * N), shape=(2 * N, 1))
        # print(k)
        A = 0.5 + N/2
        B = 0.5
        # now we have a k x n matrix
        m = (1 / N) * np.cos((np.pi / N) * (n + A) * (k + B))
        # T_square = np.array([T.tolist() for i in range(2*N)])
        # [print(line) for line in np.round(m,3).tolist()]
        imdct = np.dot(m, y)
        return imdct

    def test_sum(
            self,
            x: np.ndarray,
            k1: int,
            k2: int,
            f,
    ):
        assert len(x) % 2 == 0
        N = int(len(x) / 2)
        # n is indexed at the columns
        n = np.arange(2 * N)

        A = 0.5 + N/2
        B = 0.5
        m = f((np.pi / N) * (n + A) * (k1 + B)) * f((np.pi / N) * (n + A) * (k2 + B))
        self.logger.debug(m)
        v = np.dot(m, x) / N
        return v


class MdctUnitTest:

    def __init__(self, logger: Logging = None):
        self.logger = logger if logger is not None else logging.getLogger()

    def test(self):
        tfm = Mdct(logger=self.logger)

        x = np.random.random(100)
        self.logger.debug(x)
        y = tfm.MDCT(x=x)
        imdct = tfm.IMDCT(y=y)
        self.logger.debug(imdct)
        err = imdct - x
        abs_err = abs(err)
        assert np.sum(abs_err**2) < 0.0000000001
        return


if __name__ == '__main__':
    lgr = LoggingSingleton.get_singleton_logger(log_level=logging.DEBUG)
    obj = Mdct(logger=lgr)
    t_val = obj.test_sum(x=np.ones(10), k1=2, k2=65, f=np.sin)
    lgr.info(t_val)
    exit(0)

    MdctUnitTest(logger=lgr).test()
    exit(0)




import logging
import numpy as np
from fitxf.math.lang.measures.TextDiffInterface import TextDiffInterface
from fitxf.utils import Logging


#
# The state-of-the-art way comparing 2 texts would be to calculate the cosine distance between
# the embedding of the texts. But this requires loading a big model, and doing slow inference.
# We want something without external dependencies whatsoever, fast, simple, standalone.
#
# We implement a few low-cost, low-resource, simple measures as follows:
#
# CHARACTER DIFFERENCE or chardiff
#    Fastest, strict character dependent measure.
#    Can quickly calculate text against a list of other texts.
#    Sentence represented as a vector of its character unicode values, then appended with 0's to achieve
#    fixed length. Thus fast comparison (single matrix multiplication) can be made to a set of many reference
#    texts.
#    See function text_similarity_chardiff() below for more details.
#
#   Measure Consistency & Must Condition:
#     charfreq = 1.0 when no common character is shared (including space). e.g. 'IloveMcD' and 'ЯдюблюМзкД'
#     charfreq = 0.0 when count of characters are equal regardless of order
#
class TextDiffCharDiff(TextDiffInterface):

    APPEND_ORDINAL_DEFAULT = 0
    MODEL_PARAMS_DEFAULT = {'ref_str_len': 100, 'append_ordinal': APPEND_ORDINAL_DEFAULT}

    def __init__(
            self,
            log_time_profilings = False,
            logger = None,
    ):
        super().__init__(
            log_time_profilings = log_time_profilings,
            logger = logger,
        )
        return

    def get_model_params(self, ref_str_len: int, append_ordinal: int) -> dict:
        return {'ref_str_len': ref_str_len, 'append_ordinal': append_ordinal}

    # Represent each character by its Unicode integer, then append 0's (by default) to a fixed length
    def get_text_model(
            self,
            text,
            model_params = MODEL_PARAMS_DEFAULT,
    ):
        ref_str_len = model_params['ref_str_len']
        append_ordinal = model_params['append_ordinal']
        ordinals_list = [ord(c) for c in text[0:min(ref_str_len, len(text))]] \
                        + [append_ordinal] * max(0, (ref_str_len - len(text)))
        self.logger.debug('chardiff model for text "' + str(text) + '": ' + str(ordinals_list))
        return ordinals_list

    def text_difference(
            self,
            candidate_text,
            ref_text_list,
            # option for user to pre-calculate to whichever text model being used
            candidate_text_model = None,
            ref_text_model_list = None,
            model_params: dict = MODEL_PARAMS_DEFAULT,
            top_k = 3,
    ) -> tuple: # returns tuple of top text list & top scores list
        assert len(ref_text_list) > 0, 'No reference to compare with for text "' + str(candidate_text) + '"'

        if ref_text_model_list is not None:
            assert len(ref_text_list) == len(ref_text_model_list), \
                'Diff length text list ' + str(len(ref_text_list)) + ' model list ' + str(len(ref_text_model_list))

        id_timer = 'text_similarity (' + str(self) + ')'
        self.profiler.start_time_profiling(id=id_timer)

        res = self.text_difference_chardiff(
            candidate_text = candidate_text,
            ref_text_list = ref_text_list,
            candidate_text_model = candidate_text_model,
            ref_text_model_list = ref_text_model_list,
            model_params = model_params,
            top_k = top_k
        )

        self.profiler.record_time_profiling(
            id = id_timer,
            msg = '',
            logmsg = self.log_time_profilings,
        )
        return res

    # Measures total number of different characters with a list of reference texts, thus faster than above
    # Convert all character to unicode number, take difference, convert to 0 and 1 (non-zero), then sum each
    # row (reference text) divided by fixed string length.
    # Take the lowest value as most similar text.
    def text_difference_chardiff(
            self,
            candidate_text,
            ref_text_list,
            # Option to pass in pre-calculated text model
            candidate_text_model = None,
            ref_text_model_list = None,
            model_params: dict = MODEL_PARAMS_DEFAULT,
            top_k = 3,
    ) -> tuple: # returns tuple of top text list & top scores list
        assert len(ref_text_list) > 0, 'No reference to compare with for text "' + str(candidate_text) + '"'

        if ref_text_model_list is None:
            ref_text_model_list = [
                self.get_text_model(text=s, model_params=model_params) for s in ref_text_list
            ]
        if candidate_text_model is None:
            candidate_text_model = self.get_text_model(text=candidate_text, model_params=model_params)

        candidate_vec = np.array(candidate_text_model, dtype=np.int32)
        candidate_vec_l = np.sum(1 * (candidate_vec > 0))
        ref_matrix = np.array(ref_text_model_list, dtype=np.int32)
        # self.logger.debug(
        #     'Candidate vec:\n' + str(candidate_vec.tolist()) + '\n, ref matrix:\n'
        #     + str([v.tolist() for v in ref_matrix])
        # )
        ref_matrix_l = np.sum(1 * (ref_matrix > 0), axis=-1)
        # Each reference vector will have their new lengths, max of (candidate text, ref text)
        ref_matrix_candidate_l = np.maximum(ref_matrix_l, candidate_vec_l)
        # self.logger.debug(
        #     'Candidate vec length ' + str(candidate_vec_l) + ', ref matrix lengths ' + str(ref_matrix_l)
        #     + ', after maximum of both ' + str(ref_matrix_candidate_l)
        # )

        # Make negative differences positive, we just need absolute value
        diff = np.abs(ref_matrix - candidate_vec)
        # self.logger.debug('Diff ' + str(diff))
        # binary difference only, 0 remains 0, positive values become 1
        diff_binary = 1 * (diff > 0)
        # self.logger.debug('Binary Diff ' + str(diff_binary))
        # raise Exception('asdf')
        # By dividing by the correct max lengths of each row, we satisfy the above MUST CONDITION
        diff_res = np.sum(diff_binary, axis=-1) / ref_matrix_candidate_l
        # self.logger.debug('Measure result ' + str(diff_res))

        closest = np.argsort(diff_res)
        top_texts, top_dist = [ref_text_list[i] for i in closest], diff_res[closest].tolist()
        # self.logger.debug('Top texts ' + str(top_texts) + ', top dist ' + str(top_dist))

        return top_texts[0:min(top_k, len(top_texts))], top_dist[0:min(top_k, len(top_dist))]


class TextDiffCharDiffUnitTest:
    def __init__(self, logger=None):
        self.logger = logger if logger is not None else logging.getLogger()
        return

    def test(self):
        obj = TextDiffCharDiff(log_time_profilings=False, logger=self.logger)

        s = "privet medved"

        # Check chardiff model is correct
        AO = obj.APPEND_ORDINAL_DEFAULT
        s_rep = obj.get_text_model(
            text = s,
            model_params = {'ref_str_len': len(s)+2, 'append_ordinal': AO},
        )
        assert s_rep == [
            0x70, 0x72, 0x69, 0x76, 0x65, 0x74, 0x20, 0x6d, 0x65, 0x64, 0x76, 0x65, 0x64, AO, AO], \
            'Wrong chardiff model for "' + str(s) + '": ' + str(s_rep)

        #
        # Important test to ensure textdiff between 2 texts are the same regardless of reference fixed length
        #
        cand_txt = 'privet medved'
        # Check chardiff model again by different lengths
        for txt_compare, exp_d, ref_len in [
            # Should be no difference regardless of reference length 30 or more
            ('privet medved', 0., 30), ('privet medved', 0., 30+np.random.randint(low=50, high=100)),
            # Should be no difference regardless of reference length 30 or more
            ('privet med', 0.23076923076923078, 30),
            ('privet med', 0.23076923076923078, 30+np.random.randint(low=50, high=100)),
            # Should be no difference regardless of reference length 30 or more
            ('privet', 0.5384615384615384, 30),
            ('privet', 0.5384615384615384, 30+np.random.randint(low=50, high=100)),
            # Should be no difference regardless of reference length 30 or more
            ('', 1.0, 30), ('', 1.0, 30+np.random.randint(low=50, high=100)),
        ]:
            model_params = obj.get_model_params(
                ref_str_len = ref_len,
                append_ordinal = obj.APPEND_ORDINAL_DEFAULT,
            )
            top_texts, top_diffs = obj.text_difference(
                candidate_text = cand_txt,
                ref_text_list = [txt_compare],
                model_params = model_params,
            )
            self.logger.info(
                'chardiff for "' + str(cand_txt) + '" vs "' + str(txt_compare) + '", ref len ' + str(ref_len)
                + ': ' + str(list(zip(top_texts, top_diffs)))
            )
            assert top_diffs[0] == exp_d, \
                'chardiff for "' + str(cand_txt) + '" vs "' + str(txt_compare) + '", ref len ' + str(ref_len) \
                + ': ' + str(list(zip(top_texts, top_diffs))) + ' top diff ' + str(top_diffs[0]) + ' not ' + str(exp_d)

        # Check both metric is correct by top_k
        s_ref = 'hi how are u'
        test_text_list = [s_ref, 'how are you', 'how are', 'рш0рщц0фку0нщг']
        close_texts, close_scores = obj.text_difference(
            candidate_text = s_ref,
            ref_text_list = test_text_list,
            top_k = 4,
            model_params = obj.get_model_params(ref_str_len=50, append_ordinal=TextDiffCharDiff.APPEND_ORDINAL_DEFAULT),
        )
        self.logger.info(
            'Close texts and scores result: ' + str(list(zip(close_texts, close_scores)))
        )
        exp_scores = [0.0, 0.9166666666666666, 0.9166666666666666, 1.0]
        exp_texts = ['hi how are u', 'how are you', 'how are', 'рш0рщц0фку0нщг']
        assert close_texts == exp_texts, 'Close texts ' + str(close_texts) + ' not ' + str(exp_texts)
        assert close_scores == exp_scores,  'Close scores ' + str(close_scores) + ' not ' + str(exp_scores)

        self.logger.info('ALL TESTS PASSED')
        return


if __name__ == '__main__':
    lgr = Logging.get_default_logger(log_level=logging.INFO, propagate=False)
    TextDiffCharDiffUnitTest(logger=lgr).test()
    exit(0)



import logging
from fitxf.utils import Profiling, Logging


#
# Simple text models (no big data LLMs) for fast & convenient comparisons of text differences.
# Mainly for use in simple applications, requiring fast, simple metrics.
# The function textdiff(t1, t2):
#   Measure Consistency & Must Condition:
#     textdiff = 1.0 when totally "different" (depending on model definition) e.g. 'IloveMcD' and 'ЯдюблюМзкД'
#     textdiff = 0.0 when t1 == t2
#
class TextDiffInterface:

    def __init__(
            self,
            log_time_profilings = False,
            logger = None,
    ):
        self.log_time_profilings = log_time_profilings
        self.logger = logger if logger is not None else logging.getLogger()
        self.profiler = Profiling(logger=self.logger)
        return

    def get_model_params(self, **kwargs) -> dict:
        return {k: v for k, v in kwargs.items()}

    def get_text_model(
            self,
            text: str,
            model_params = {},
    ):
        raise Exception('Must be implemented by child class!!')

    def text_difference(
            self,
            candidate_text,
            ref_text_list,
            # option for user to pre-calculate to whichever text model being used
            candidate_text_model = None,
            ref_text_model_list = None,
            model_params = {},
            top_k = 3,
    ) -> tuple: # returns tuple of top text list & top scores list
        raise Exception('Must be implemented by child class!!')

    def text_similarity(
            self,
            candidate_text,
            ref_text_list,
            # option for user to pre-calculate to whichever text model being used
            candidate_text_model = None,
            ref_text_model_list = None,
            model_params = {},
            top_k = 3,
    ) -> tuple: # returns tuple of top text list & top scores list
        top_texts, top_scores = self.text_difference(
            candidate_text = candidate_text,
            ref_text_list = ref_text_list,
            candidate_text_model = candidate_text_model,
            ref_text_model_list = ref_text_model_list,
            model_params = model_params,
            top_k = top_k,
        )
        return [t for t in reversed(top_texts)], [v for v in reversed(top_scores)]


if __name__ == '__main__':
    lgr = Logging.get_default_logger(log_level=logging.INFO, propagate=False)
    exit(0)



import logging
import numpy as np
import matplotlib.pyplot as plt
from fitxf.utils import Logging
from fitxf.math.utils.LoggingSingleton import LoggingSingleton


class Dft:

    def __init__(self, logger: Logging = None):
        self.logger = logger if logger is not None else logging.getLogger()
        return

    def getSineSignal(
            self,
            amps: np.ndarray,
            freqs: np.ndarray,
            samp_rate: float,
            show_plot: bool = False,
    ):
        assert len(amps) == len(freqs)

        dt = 1.0 / samp_rate
        t_start = 0
        t_end = 1
        t = np.arange(t_start, t_end, dt)

        x = np.zeros_like(t)
        for a, f in zip(amps, freqs):
            x += a * np.sin(2 * np.pi * f * t)

        if show_plot:
            plt.figure(figsize=(8, 6))
            plt.plot(t, x, 'r')
            plt.xlabel('t')
            plt.ylabel('x')
            plt.grid(True)
            plt.show()

        return x

    def DFT(
            self,
            # time domain signal
            x: np.ndarray,
            lib: str = 'dot'
    ):
        if lib == 'numpy':
            return np.fft.fft(x)
        else:
            N = x.size
            # n is indexed at the columns
            n_vec = np.arange(N)
            # reshape so that k is indexed at the rows
            k_vec = n_vec.reshape((N, 1))
            # print(k_vec)
            # raise Exception('asdf')

            e = np.exp(-2j * np.pi * k_vec * n_vec / N)
            # print(e, e.shape)
            # raise Exception('asdf')
            return np.dot(e, x)

    def IDFT(
            self,
            y: np.ndarray,
    ):
        N = y.size
        # k is indexed at the columns
        k_vec = np.arange(N)
        # reshape so that n is indexed at the rows
        n_vec = k_vec.reshape((N, 1))
        # print(k_vec)
        # raise Exception('asdf')

        e = np.exp(2j * np.pi * k_vec * n_vec / N)
        # print(e, e.shape)
        # raise Exception('asdf')
        return (1 / N) * np.dot(e, y)

    def plot_fft(self, X_fft: np.ndarray, freq: np.ndarray):
        plt.figure(figsize=(8,6))
        plt.stem(freq, abs(X_fft), 'b', markerfmt=' ', basefmt='-b')
        plt.xlabel(r'$\omega$ '+'(Hz)')
        plt.ylabel('DFT Amplitude '+r'$|X(\omega)|$')
        plt.grid(True)
        plt.show()
        return


class DftUnitTest:

    def __init__(self, logger: Logging = None):
        self.logger = logger if logger is not None else logging.getLogger()

    def test(self):
        dft = Dft(logger=self.logger)

        x = np.random.random(100)

        y = dft.DFT(x=x)
        idft = dft.IDFT(y=y)
        err = idft - x
        abs_err = abs(err)
        assert np.sum(abs_err**2) < 0.0000000001
        return


if __name__ == '__main__':
    lgr = LoggingSingleton.get_singleton_logger()
    DftUnitTest(logger=lgr).test()

    # sampling rate 100Hz
    rate = 30

    dft = Dft()

    amps = np.array([1, 2, 10])
    freqs = np.array([1, 2, 10])
    x = dft.getSineSignal(
        samp_rate=rate,
        amps=amps,
        freqs=freqs,
    )
    lgr.info('x shape ' + str(x.shape) + ': ' + str(x))

    X_k = dft.DFT(x=x)
    lgr.info('DFT ' + str(abs(X_k)) + ', shape ' + str(X_k.shape))
    N = X_k.size
    n = np.arange(N)
    # total time span
    T = N / rate
    freq = n / T
    lgr.info('N=' + str(N) + ', total time span T=' + str(T))
    lgr.info('Discrete freq: ' + str(freq))

    dft.plot_fft(X_k, freq=freq)
    idft = dft.IDFT(y=X_k)
    lgr.info('Inverse DFT error ' + str(idft - x))
    exit(0)




import logging
import numpy as np
from fitxf.math.lang.measures.TextDiffInterface import TextDiffInterface
from fitxf.math.lang.measures.TextDiffCharDiff import TextDiffCharDiff
from fitxf.math.lang.measures.TextDiffCharFreq import TextDiffCharFreq
from fitxf.utils import Profiling, Logging


class TextDiffSpeedTest:
    def __init__(self, logger=None):
        self.logger = logger if logger is not None else logging.getLogger()
        return

    def test(self):
        td_chardiff = TextDiffCharDiff(log_time_profilings=False, logger=self.logger)
        td_charfreq = TextDiffCharFreq(log_time_profilings=False, logger=self.logger)

        for td, itr, exp_rps, n_ref in [
            # Char freq needs to loop 1-by-1 in n ref sentences, thus the expected RPS is super small
            (td_charfreq, 50, 8.0, 1000),
            # Char diff algo is fast with high expected RPS
            (td_chardiff, 2000, 280.0, 1000),
        ]:
            self.logger.info(
                'Start speed test for ' + str(td) + ', n ref ' + str(n_ref) + ' exp rps ' + str(exp_rps)
                + ', iterations ' + str(itr) + '...'
            )
            self.test_speed(
                td = td,
                itr = itr,
                exp_rps = exp_rps,
                n_ref = n_ref,
            )
        self.logger.info('SPEED TESTS PASSED')
        return

    def test_speed(
            self,
            td: TextDiffInterface,
            itr: int,
            exp_rps: float,
            # how many alphabets in our fake language
            num_lang_chars: int = 88,
            min_sentence_len: int = 50,
            max_sentence_len: int = 150,
            # how many sentences in reference text list
            n_ref: int = 1000,
    ):
        prf = Profiling(logger=self.logger)
        model_prms = {'ref_str_len': max_sentence_len, 'append_ordinal': 0} if type(td) in [TextDiffCharDiff] \
            else {'n_tuple': 3}
        l_avg = int((min_sentence_len + max_sentence_len) / 2)
        alphabets = [' '] \
                    + [chr(x) for x in range(ord('a'), ord('z')+1, 1)]
                    # + [chr(x) for x in range(ord('A'), ord('Z')+1, 1)]
                    # + [chr(x) for x in range(ord('0'), ord('9')+1, 1)] \
                    # + [chr(x) for x in range(ord('а'), ord('я')+1, 1)] \
                    # + [chr(x) for x in range(ord('А'), ord('Я')+1, 1)]
        # self.logger.info('Alphabets: ' + str(alphabets))

        # random text lengths between min_sentence_len to max_sentence_len
        rl = np.round(l_avg + (l_avg * np.random.rand(n_ref) - l_avg/2), decimals=0).astype(np.int32)
        # self.logger.info('Random text lengths ' + str(rl) + ', min ' + str(np.min(rl)) + ', max ' + str(np.max(rl)))

        s_ref = ''.join(np.random.choice(alphabets, l_avg).tolist())
        s_ref_model = td.get_text_model(text=s_ref, model_params=model_prms)
        # self.logger.info('String ref length ' + str(len(s_ref_model)) + ': ' + str(s_ref_model))
        ref_text_list = [
            ''.join(np.random.choice(alphabets, s_len).tolist())
            for s_len in rl
        ]
        ref_text_model_list = [
            td.get_text_model(text=s, model_params=model_prms) for s in ref_text_list
        ]
        # self.logger.info('Ref text list ' + str(ref_text_list))
        # self.logger.info('Ref text list model lengths ' + str([len(v) for v in ref_text_model_list]))

        # Test speed, make sure below some threshold
        self.logger.info('Started speed test for ' + str(td))
        start_time = prf.start()
        for i in range(itr):
            a, b = td.text_similarity(
                candidate_text = s_ref,
                ref_text_list = ref_text_list,
                candidate_text_model = s_ref_model,
                ref_text_model_list = ref_text_model_list,
                top_k = 4,
                model_params = model_prms,
            )
            # self.logger.info(str(a) + ': ' + str(b))
            # raise Exception('asdf')

        diff_secs = prf.get_time_dif_secs(start=start_time, stop=prf.stop(), decimals=8)
        rps = round(itr / diff_secs, 2)
        if rps < exp_rps:
            self.logger.warning(
                'Test class ' + str(type(td)) + ', ' +  str(itr) + ' iterations, FAIL rps = ' + str(rps)
                + ' < ' + str(exp_rps)
            )
        else:
            self.logger.info(
                'Test class ' + str(type(td)) + ', ' +  str(itr) + ' iterations, OK rps = ' + str(rps)
                + ' >= ' + str(exp_rps)
            )
        return


if __name__ == '__main__':
    lgr = Logging.get_default_logger(log_level=logging.INFO, propagate=False)
    TextDiffSpeedTest(logger=lgr).test()
    exit(0)



import logging
import collections
import numpy as np
from fitxf.math.lang.measures.TextDiffInterface import TextDiffInterface
from fitxf.utils import Logging


#
# The state-of-the-art way comparing 2 texts would be to calculate the cosine distance between
# the embedding of the texts. But this requires loading a big model, and doing slow inference.
# We want something without external dependencies whatsoever, fast, simple, standalone.
#
# We implement a few low-cost, low-resource, simple measures as follows:
#
# 1. CHARACTER FREQUENCY or "charfreq"
#   Slower than chardiff, non-strict character position independent measure.
#   Cannot quickly calculate between a text and a list of texts.
#   Fast implementation of how different are 2 texts. Min 0 (no diff), max 1.0 (100% different)
#   In a way it is using similar concepts along the lines of "Bog of Words", but is "Bag of Characters"
#   instead. Other differences include we don't maintain a vocabulary of characters and do not use
#   cosine similarity.
#   Instead, we implement a totally standalone, no admin management embedding models, vocab,
#   simple & convenient measure, purely between 2 texts.
#
#   Mathematically between 2 texts, charfreq is given by
#     SUM {abs(count_candidate_text(c) - count_ref_text(c))} / SUM{max(character_count(c, candidate_text, ref_text))}
#   over all characters c in both candidate_text and text2
#
#   Measure Consistency & Must Condition:
#     charfreq = 1.0 when no common character is shared (including space). e.g. 'IloveMcD' and 'ЯдюблюМзкД'
#     charfreq = 0.0 when count of characters are equal regardless of order
#
class TextDiffCharFreq(TextDiffInterface):

    MODEL_PARAMS_DEFAULT = {'n_tuple': 3}

    def __init__(
            self,
            log_time_profilings = False,
            logger = None,
    ):
        super().__init__(
            log_time_profilings = log_time_profilings,
            logger = logger,
        )
        return

    def get_model_params(self, n_tuple: int) -> dict:
        return {'n_tuple': n_tuple}

    # Just a simple dictionary of tuple character frequencies
    # E.g. "privet medved", n tuple 3 becomes
    #   {'pri': 1, 'riv': 1, 'ive': 1, 'vet': 1, 'et ': 1, 't m': 1, ' me': 1, 'med': 1, 'edv': 1, 'dve': 1}
    # For n tuple = 2
    #   {'pr': 1, 'ri': 1, 'iv': 1, 've': 2, 'et': 1, 't ': 1, ' m': 1, 'me': 1, 'ed': 1, 'dv': 1}
    # For n tuple = 1
    #   {'p': 1, 'r': 1, 'i': 1, 'v': 2, 'e': 3, 't': 1, ' ': 1, 'm': 1, 'd': 1}
    def get_text_model(
            self,
            text: str,
            model_params = MODEL_PARAMS_DEFAULT,
    ):
        n_tuple = model_params['n_tuple']
        s_consecutive_chars = [text[i:(i+n_tuple)] for i in range(len(text)) if i<len(text) - n_tuple]
        counter_text_model = dict(collections.Counter(s_consecutive_chars))
        self.logger.debug(
            'charfreq model for text "' + str(text) + '", n tuple ' + str(n_tuple) + ': ' + str(counter_text_model)
        )
        return counter_text_model

    def text_difference(
            self,
            candidate_text,
            ref_text_list,
            # option for user to pre-calculate to whichever text model being used
            candidate_text_model = None,
            ref_text_model_list = None,
            model_params: dict = MODEL_PARAMS_DEFAULT,
            top_k = 3,
    ) -> tuple: # returns tuple of top text list & top scores list
        assert len(ref_text_list) > 0, 'No reference to compare with for text "' + str(candidate_text) + '"'

        if ref_text_model_list is not None:
            assert len(ref_text_list) == len(ref_text_model_list), \
                'Diff length text list ' + str(len(ref_text_list)) + ' model list ' + str(len(ref_text_model_list))

        id_timer = 'text_similarity (' + str(self) + ')'
        self.profiler.start_time_profiling(id=id_timer)

        candidate_counter = self.get_text_model(text=candidate_text, model_params=model_params)
        diff_measures = []
        for i, ref_text in enumerate(ref_text_list):
            diff = self.text_difference_charfreq(
                candidate_text_or_model = candidate_counter if candidate_text_model is None else candidate_text_model,
                ref_text_or_model = ref_text if ref_text_model_list is None else ref_text_model_list[i],
                model_params = model_params,
            )
            diff_measures.append(diff)

        diff_measures = np.array(diff_measures)
        closest = np.argsort(diff_measures)
        top_texts, top_dist = [ref_text_list[i] for i in closest], diff_measures[closest].tolist()
        res = top_texts[0:min(top_k, len(top_texts))], top_dist[0:min(top_k, len(top_dist))]

        self.profiler.record_time_profiling(
            id = id_timer,
            msg = '',
            logmsg = self.log_time_profilings,
        )
        return res

    # Mathematically between 2 texts, charfreq is given by
    #   SUM {abs(count_candidate_text(c) - count_ref_text(c))} / SUM{max(character_count(c, candidate_text, ref_text))}
    # over all characters c in both candidate_text and text2
    def text_difference_charfreq(
            self,
            candidate_text_or_model,
            ref_text_or_model,
            model_params: dict = MODEL_PARAMS_DEFAULT,
    ):
        toc = candidate_text_or_model
        counter_candidate = self.get_text_model(text=toc, model_params=model_params) if type(toc) is str else toc
        toc = ref_text_or_model
        counter_ref = self.get_text_model(text=toc, model_params=model_params) if type(toc) is str else toc

        all_unique_keys = list(set(list(counter_candidate.keys()) + list(counter_ref.keys())))
        all_unique_keys_max = {
            k: max(
                counter_candidate.get(k, 0),
                counter_ref.get(k, 0)
            )
            for k in all_unique_keys
        }
        # sum the total of the max frequencies of all tuples
        total_count = np.array([all_unique_keys_max[k] for k in all_unique_keys])
        # self.logger.debug('Total count ' + str(total_count) + ' from map ' + str(all_unique_keys_max))
        diff = [
            abs(
                counter_candidate.get(k, 0) - counter_ref.get(k, 0)
            )
            for k in all_unique_keys
        ]
        np_diff = np.array(diff)
        # self.logger.debug(
        #     'keys:\n'+ str(all_unique_keys) + '\nnp diff:\n' + str(np_diff) + '\ntotal counts:\n' + str(total_count)
        # )
        # diff_map = [(k, v) for k,v in zip(all_unique_keys, diff)]
        diff_measure = np.sum(np_diff) / np.sum(total_count)
        self.logger.debug(
            'Difference measure = ' + str(diff_measure) # + ', difference map ' + str(diff_map)
            + ', keys max map ' + str(all_unique_keys_max)
            + ' for texts "' + str(candidate_text_or_model) + '" and "' + str(ref_text_or_model) + '"'
        )
        return diff_measure


class TextDiffCharFreqUnitTest:
    def __init__(self, logger=None):
        self.logger = logger if logger is not None else logging.getLogger()
        return

    def test(self):
        obj = TextDiffCharFreq(log_time_profilings=False, logger=self.logger)

        s = "privet medved"

        # Check charfreq model is correct
        s_rep = obj.get_text_model(text=s, model_params={'n_tuple': 2})
        assert s_rep == {'pr': 1, 'ri': 1, 'iv': 1, 've': 2, 'et': 1, 't ': 1, ' m': 1, 'me': 1, 'ed': 1, 'dv': 1}, \
            'Wrong charfreq model for "' + str(s) + '": ' + str(s_rep)

        # Check charfreq metric is correct
        for t1, t2, expected_diff in (
                # Exactly similar text
                ('exactly same text must be 0.0 measure', 'exactly same text must be 0.0 measure', 0.0),
                # Completely different text
                ('IloveMcD', '난는맥디를좋아', 1.0),
                # Usual cases
                ('1234567890', '12345abcde', 0.615385),
                ('your facebook OTP is 123456', 'your facebook OTP is 123666', 0.142857),
                # diff measure will not be 100% here because of common space ' ' characters
                ('I love McD', '난는 맥디를 좋아', 0.866667),
        ):
            diff = obj.text_difference_charfreq(
                candidate_text_or_model = t1,
                ref_text_or_model = t2,
                model_params = {'n_tuple': 1},
            )
            self.logger.info(
                'Charfreq between "' + str(t1) + '" & "' + str(t2) + '": ' + str(diff)
            )
            assert round(diff, 6) == expected_diff, \
                'Diff ' + str(diff) + ' not equal to expected diff ' + str(expected_diff)

        # Check both metric is correct by top_k
        test_text_list = ['test test', 'hi how are u', 'how are you', '****asdf 0123456789 0123456789', 'how are']
        close_texts, close_scores = obj.text_difference(
            candidate_text = 'hi how are you',
            ref_text_list = test_text_list,
            top_k = 4,
            model_params = {'n_tuple': 1},
        )
        # exp_scores = [0.14285714285714285, 0.21428571428571427, 0.5, 0.9047619047619048]
        exp_scores = [0.15384615384615385, 0.23076923076923078, 0.5384615384615384, 0.8947368421052632]
        assert close_texts == ['hi how are u', 'how are you', 'how are', 'test test'], \
            'Close texts ' + str(close_texts)
        assert close_scores == exp_scores,  'Close scores ' + str(close_scores) + ' not ' + str(exp_scores)

        self.logger.info('ALL TESTS PASSED')
        return


if __name__ == '__main__':
    lgr = Logging.get_default_logger(log_level=logging.DEBUG, propagate=False)
    TextDiffCharFreqUnitTest(logger=lgr).test()
    exit(0)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# https://github.com/coderByNeed/BOB404/blob/master/audio/compressions.py
from __future__ import absolute_import, division, print_function, with_statement
from sys import path
from os.path import dirname, abspath

path.append(dirname(abspath(__file__)))

# --------------------------------------------------------------------------------------------------#
# -----------------------------------------------HEAD-----------------------------------------------#
# --------------------------------------------------------------------------------------------------#


## WAVE FILE IMPORTING------------------------------------------------------------------------------
from struct import pack, unpack

## WAVE FILE CLEANUP -------------------------------------------------------------------------------
##>> METADATA
__author__ = 'Danilo Mitrovic ,'
__credits__ = ''
__copyrights__ = ''
__licence__ = 'GNU'
__version__ = '1.01'
__maintainer__ = __author__
__email__ = 'adewdda@gmail.com'
__status__ = 'Development'
__contact__ = 'not yet'

## WAVE FILE GLOBAL VARIABLES-----------------------------------------------------------------------
BITMASK = 1
NIBMASK = 0XF
BYTMASK = 0XFF

NIBSHIFT = 4
BYTSHIFT = 8


## WAVE FILE DEF FUNCTIONS--------------------------------------------------------------------------

def least_sagnificant_bit(int_numb):  ##returns position of the bit not the value of the bit
    if int_numb == 0: return 0
    pos = 0
    while 1:
        if int_numb & (BITMASK << pos) != 0: return pos + 1
        pos += 1


def most_sagnificant_bit(int_numb):
    BYTSHIFT = 8
    mask = 0x8000000000000000
    chk_msk = 0xff00000000000000
    pos = 64
    while 1:

        if int_numb & chk_msk == 0:
            chk_msk >>= BYTSHIFT
            mask >>= BYTSHIFT
            pos -= BYTSHIFT
        else:

            if int_numb & mask == 0:
                mask >>= BITMASK
                pos -= BITMASK
            else:
                break

        if pos == 0: break
    return pos


def bytemask(int_numb, mask=None):
    msk = 0
    i = 0
    while 1:

        msk += BYTMASK << (i * 8)

        if int_numb & msk == int_numb:
            return msk
        i += 1


def twos_compliment(int_numb):
    if not int_numb: return ~int_numb

    lsb = least_sagnificant_bit(int_numb)
    msb = most_sagnificant_bit(int_numb)
    mask = (1 << lsb) - 1
    mask ^= (1 << msb) - 1
    return int_numb ^ mask


## WAVE FILE GLOBAL FUNCTIONS-----------------------------------------------------------------------

# :D unpacking
u_signed_8bit_l = lambda *x: unpack('<' + 'h' * len(x), *x) if len(x) == 1 \
    else unpack('<' + 'h' * len(x), *x)[0]
u_unsgnd_8bit_l = lambda *x: unpack('<' + 'H' * len(x), *x) if len(x) == 1 \
    else unpack('<' + 'H' * len(x), *x)[0]
u_unsgnd_16bt_l = lambda *x: unpack('<' + 'I' * len(x), *x) if len(x) == 1 \
    else unpack('<' + 'I' * len(x), *x)[0]

# :D packing
p_signed_8bit_l = lambda *x: pack('<' + 'h' * len(x), *x) if len(x) == 1 \
    else pack('<' + 'h' * len(x), *x)[0]
p_unsgnd_8bit_l = lambda *x: pack('<' + 'H' * len(x), *x) if len(x) == 1 \
    else pack('<' + 'H' * len(x), *x)[0]
p_unsgnd_16bt_l = lambda *x: pack('<' + 'I' * len(x), *x) if len(x) == 1 \
    else pack('<' + 'I' * len(x), *x)[0]

# :D byte conversion
bit_to_int = lambda x: ord(x) if type(x) != int else x

checkNIB = lambda x: '{0:04b}'.format(x) if x else '0000'
checkBYT = lambda x: '-'.join(map(checkNIB, [(x >> 4) & 0xf, x & 0xf]))
checkBTS = lambda x: ' '.join([checkBYT((x >> (shift * 8)) & 0xff) for shift in reversed(range(4))])


def hex_byte(int_d):
    int_d = checkBYT(int_d) if isinstance(int_d, int) else int_d
    int_d = int_d.split('-')
    int_d = map(lambda x: int(x, 2), int_d)
    int_d = map('{0:X}'.format, int_d)
    return ''.join(int_d)


def hex_bytes(int_d):
    return '-'.join(map(hex_byte, checkBTS(int_d).split(' ')))


####################################################################################################
################################################MAIN################################################
####################################################################################################

## :D G.711 µ LAw
# :D MULAW decoding from 8bit integer to 16bit integer
def u_law_d(i8bit):
    i8bit &= 0xff  # marginalising data larger than byte
    i8bit ^= 0xff  # flipping back bytes
    sign = False

    if i8bit & 0x80 == 0x80:  # if it is signed negative 1000 0000
        sign = True  # bool option since sign is not really used
        i8bit &= 0x7f  # removing the sign of value

    pos = ((i8bit & 0xf0) >> 4) + 5  # grabing initial value for mantisa

    # generating decoded data
    decoded = i8bit & 0xf  # grabing 1st nibble from 8 bit integer
    decoded <<= pos - 4  # shifting by position -4 aka generating mantisa for 16 bit integer
    decoded |= 1 << (pos - 5)  # OR gate for specific bit
    decoded |= 1 << pos  # OR gate for another specific bit
    decoded -= 0x21  # removing the 10 0001 from value to generate exact value

    if not sign:
        return decoded  # if positive number will be returned as is
    else:
        return -decoded  # if negative the number will be returned inverted


# :D MULAW encoding from 16bit integer to 8bit integer
def u_law_e(i16bit):
    i16bit &= 0x3fff  # strips data bigger than 14 bits
    pos = 12  # position is 12 since we are not calculating sign bit and 0 value so 14 - sign == 13; so 13 bits, - 0 value ... 12 bits
    msk = 0x1000  # mask 1 0000 0000 0000 # bitwise mask for singular bits

    sign = 0x80 if i16bit & 0x2000 else 0  # generating 8bit sign from 14 bit number

    if sign == 0x80:  # if sign is present
        i16bit = twos_compliment(i16bit)  # twos_compliment
        i16bit &= 0x3ffe  # strips all bits larger than 14 bits

    i16bit += 0b100001  # adds 33, 0x21 = 10 0001

    if i16bit > 0x1fff: i16bit = 0x1fff  # if number is over maximum ... it becomes maximum

    for x in reversed(
            range(pos)):  # this number has least significant bits, not bit... so they have often size of 4 bits, that is why it must be larger than 5
        if (i16bit & msk) != msk and pos >= 5:
            pos = x
            msk >>= 1

    LSBTS = (i16bit >> (pos - 4)) & 0xf  # grabbing mantissa from 16 bit integer

    encoded = sign  # sign
    encoded += (pos - 5) << 4  # exponent
    encoded += LSBTS  # mantisa

    return encoded ^ 0xff  # inverting all bits


ulaw_Data = [-0x1fdf, -0xfef, -0x0001, 0x0000, 0x0001, 0xfef, 0x1fdf]

form = lambda x: '{0}\t:>\t{1}\t||\t{2}'.format(x, hex_bytes(x), checkBTS(x))

print('-' * 120)
print('-' * 120)

for data in ulaw_Data:
    print('')

    data &= 0x3fff
    print('normal :\t' + form(data) + ' (' + str(type(data)) + ' ' + str(data) + ')')

    e_data = u_law_e(data)
    print('encoded:\t' + form(e_data) + ' (' + str(type(e_data)) + ' ' + str(e_data) + ')')

    d_data = u_law_d(e_data)
    print('decoded:\t' + form(d_data) + ' (' + str(type(d_data)) + ' ' + str(d_data) + ')')

    print('-' * 120)

print('-' * 120)



import logging
import numpy as np
import math
from fitxf.math.lang.cache.TextSimilarityCache import TextSimilarityCache
from fitxf.utils import Logging, Profiling


class CacheTest:

    def __init__(self, logger: Logging = None):
        self.logger = logger if logger is not None else logging.getLogger()
        self.prf = Profiling(logger=self.logger)
        return

    def test(self):
        ref_texts = [
            'The Chinese ate the dimsum by the street in Beijing',
            'The American gulped down the burger',
            'The French was buying a croissant in the local bakery by the river',
            'The Korean bought a pack of kimchi',
            'The Japanese paid for ramen with a 1000 yen bill to the robot machine',
            'The Pinoys loved their local Inasal chicken with Dinuguang',
            'The Honkies sell very expensive dimsum and the aircons keep dripping water by the walk pavements',
        ] * math.floor(1000 / 7)
        random_numbers = np.round(np.random.rand(len(ref_texts)), decimals=10).tolist()
        ref_texts = [t+' '+str(r) for t, r in zip(ref_texts, random_numbers)]
        test_texts = [
            'Tomorrow is a holiday',
            'The Chinese ate the dimsum by the street',
        ] * 100

        cache = TextSimilarityCache(
            cache_name = 'unit_test',
            cache_size = 1000,
            clear_cache_method = 'old',
            rm_prop_when_full = 0.5,
            text_similarity_fixed_len = 100,
            logger = self.logger,
        )

        for t in ref_texts:
            cache.add_to_cache_threadsafe(object=t, result={'txt': t, 'len': len(t)})
        self.logger.info(
            'Added to cache total objects ' + str(len(ref_texts)) + ', cache size ' + str(cache.get_cache_size())
        )

        start_time = self.prf.start()
        diff_thr = 0.4
        hits = 0
        misses = 0
        for t in test_texts:
            res = cache.get_from_cache_threadsafe(object=t, difference_threshold=diff_thr)
            if res:
                hits += 1
            else:
                misses += 1
            # self.logger.info('For object "' + str(t) + '", dif thr ' + str(diff_thr) + ' got ' + str(res))
        diff_secs = self.prf.get_time_dif_secs(start=start_time, stop=self.prf.stop())
        rps = round(len(ref_texts) / diff_secs, 2)
        self.logger.info(
            'Tested get from cache total ' + str(len(test_texts)) + ', took ' + str(diff_secs)
            + 's, rps ' + str(rps) + ' per sec. Hits ' + str(hits) + ', misses ' + str(misses)
        )
        assert hits == misses == 100
        return


if __name__ == '__main__':
    lgr = Logging.get_default_logger(log_level=logging.INFO, propagate=False)
    CacheTest(lgr).test()
    exit(0)



import logging
import threading
import os
import pandas as pd
from fitxf.math.lang.measures.TextDiffCharDiff import TextDiffCharDiff
from datetime import datetime, timedelta
from fitxf import DbParams
from fitxf import Datastore as DatastoreMaster
from fitxf.utils import Logging, Profiling, Env


#
# A key/value cache (bad naming below makes it "object"/"result")
#
class TextSimilarityCache:

    KEY_ID = 'key'   # key
    KEY_VALUE = 'value'   # value
    KEY_OBJECT_ORI = 'original_object'
    KEY_REPEAT_COUNT = 'repeat_count'
    KEY_DATETIME = 'datetime'

    DATETIME_FORMAT = '%Y-%m-%d %H:%M:%S.%f'

    MAX_KEY_LEN = 128
    CREATE_TABLE_BY_DB_TYPE = {
        'csv': '',
        'mysql':  """CREATE TABLE `<TABLENAME>` (
               `""" + str(KEY_ID) + """` varchar(""" + str(MAX_KEY_LEN) + """) NOT NULL,
               `""" + str(KEY_VALUE) + """` TEXT NOT NULL,
               `""" + str(KEY_DATETIME) + """` varchar(64) NOT NULL,
               `""" + str(KEY_OBJECT_ORI) + """` TEXT NOT NULL,
               PRIMARY KEY (`""" + str(KEY_ID) +  """`)
            )""",
    }

    def __init__(
            self,
            cache_name: str = None,
            cache_size: int = 1000,
            db_params: DbParams = None,
            # Below variables only for dict type cache
            # allowed values 'old', 'unpopular'
            clear_cache_method: str = 'old',
            # How much to remove from unproven cache when full, default is 50%
            rm_prop_when_full: float = 0.5,
            text_similarity_fixed_len: int = 100,
            logger: Logging = None,
    ):
        self.cache_name = cache_name
        self.cache_size = cache_size
        assert self.cache_size > 0
        self.db_params = db_params
        self.clear_cache_method = clear_cache_method
        assert self.clear_cache_method in ['old', 'unpopular']
        self.rm_prop_when_full = rm_prop_when_full
        self.text_similarity_fixed_len = text_similarity_fixed_len
        self.logger = logger if logger is not None else logging.getLogger()

        self.cache_dict = {}
        if self.db_params is not None:
            self.db_params.db_create_table_sql = self.CREATE_TABLE_BY_DB_TYPE[self.db_params.db_type]
            self.cache_db = DatastoreMaster(db_params=db_params, logger=self.logger).get_data_store()
            self.logger.info('Successfully created cache db from db params: ' + str(self.db_params.get_db_info()))
        self.cache_stats = {'name': self.cache_name, 'total': 0, 'hit_exact': 0, 'hit_similar': 0, 'hit_rate': 0.0}

        self.textdiff = TextDiffCharDiff(
            log_time_profilings = True,
            logger = self.logger,
        )
        self.textdiff_model_prms = self.textdiff.get_model_params(
            ref_str_len = self.text_similarity_fixed_len,
            append_ordinal = 0,
        )
        self.ref_texts_keys = []
        self.ref_texts_chardiff_model = []

        self.__mutex_cache = threading.Lock()
        self.profiler = Profiling(logger=self.logger)

        self.logger.info(
            'Initialized SimpleCache with cache size ' + str(self.cache_size)
            + ', clear cache method "' + str(self.clear_cache_method)
            + '", remove proportion when cache full ' + str(self.rm_prop_when_full)
        )
        return

    def get_cache_size(self):
        try:
            self.__mutex_cache.acquire()
            return len(self.cache_dict)
        finally:
            self.__mutex_cache.release()

    def update_cache_stats(
            self,
            hit_exact = 0,
            hit_similar = 0,
    ):
        assert hit_exact + hit_similar <= 1, 'Hit exact ' + str(hit_exact) + ', hit similar ' + str(hit_similar)
        # Total always add 1
        self.cache_stats['total'] += 1
        self.cache_stats['hit_exact'] += hit_exact
        self.cache_stats['hit_similar'] += hit_similar

        if self.cache_stats['total'] > 0:
            self.cache_stats['hit_rate'] = \
                (self.cache_stats['hit_exact'] + self.cache_stats['hit_similar']) / self.cache_stats['total']
        # Log stats every now and then
        if self.cache_stats['total'] % 500 == 0:
            self.logger.info('Key Value Cache "' + str(self.cache_name) + '" stats now: ' + str(self.cache_stats))
        return

    def clear_cache(self):
        if self.db_params is not None:
            self.cache_db.delete_all(
                # key = self.KEY_ID,
                tablename = self.db_params.db_table,
            )
            self.logger.info('Cleared cache for db params: ' + str(self.db_params.get_db_info()))
        else:
            self.cache_dict = {}
            self.logger.info('Cleared cache of dict type.')
        return

    def get_cache_keys(self):
        if self.db_params is not None:
            res = self.cache_db.get_all(
                tablename = self.db_params.db_table,
                max_records = 1000000,
            )
            df_tmp = pd.DataFrame.from_records(res)
            df_tmp.sort_values(by=[self.KEY_DATETIME], ascending=True, inplace=True)
            ckeys = [r[self.KEY_ID] for r in df_tmp.to_dict(orient='records')]
            return ckeys
        else:
            return list(self.cache_dict.keys())

    def derive_key(
            self,
            object,
    ):
        obj_key = str(object)
        if len(obj_key) > self.MAX_KEY_LEN:
            # obj_key = hashlib.md5(obj_key.encode(encoding="utf-8")).hexdigest()
            obj_key = obj_key[:self.MAX_KEY_LEN]
        return obj_key

    # Look for <object> and return RESULT, something like lookup "key" and return "value", just naming problem
    def get_from_cache_threadsafe(
            self,
            object,
            # by default is exact key search. if this value >0, means will do a text difference search
            difference_threshold = 0.0,
    ):
        if self.db_params is not None:
            # Clear records older than 1 day
            dt = datetime.now() - timedelta(days=1)
            dt_str = dt.strftime(self.DATETIME_FORMAT)
            del_sql = "DELETE FROM `" + str(self.db_params.db_table) + "` WHERE `" \
                        + str(self.KEY_DATETIME) + "` <= '" + str(dt_str) + "'"
            try:
                res = self.cache_db.delete_by_raw_query(
                    raw_query = del_sql,
                    tablename = self.db_params.db_table,
                )
                self.logger.info('Cleanup cache delete query "' + str(del_sql) + '": ' + str(res))
            except Exception as ex:
                self.logger.error('Failed to delete using sql "' + str(del_sql) + '": ' + str(ex))
                pass

            key = self.derive_key(object=object)
            res = self.cache_db.get(
                match_phrase = {self.KEY_ID: key},
                tablename = self.db_params.db_table,
            )
            self.logger.info(
                'Fetch from cache key "' + str(key) + '": ' + str(res)
            )
            if len(res) == 1:
                return res[0][self.KEY_VALUE]
            else:
                return None
        else:
            return self.get_from_cache_dict_threadsafe(
                object = object,
                difference_threshold = difference_threshold,
            )

    # Look for <object> and return RESULT, something like lookup "key" and return "value", just naming problem
    def get_from_cache_dict_threadsafe(
            self,
            object,
            # by default is exact key search. if this value >0, means will do a text similarity search
            difference_threshold = 0.0,
    ):
        key = self.derive_key(object=object)
        hit_exact, hit_similar = 0, 0

        try:
            self.__mutex_cache.acquire()

            if self.db_params is not None:
                res = self.cache_db.get(
                    match_phrase = {self.KEY_ID: key},
                    tablename = self.db_params.db_table,
                )
                if len(res) > 0:
                    self.logger.info(
                        'Found exact match in cache "' + str(self.db_params.db_type) + '" for key "' + str(key)
                        + '": ' + str(res)
                    )
                    hit_exact = 1
                else:
                    return None
            else:
                # First we try to look for exact match in proven cache
                if key in self.cache_dict.keys():
                    self.cache_dict[key][self.KEY_REPEAT_COUNT] += 1
                    self.logger.info(
                        'Found exact match in cache "' + str(key) + '": ' + str(self.cache_dict[key])
                        + ' Text repeat count now ' + str(self.cache_dict[key][self.KEY_REPEAT_COUNT])
                        + ' "' + str(key) + '"'
                    )
                    hit_exact = 1
                    return self.cache_dict[key][self.KEY_VALUE]
                elif difference_threshold > 0.0:
                    if not self.ref_texts_keys:
                        return None
                    top_keys, top_distances = self.textdiff.text_difference(
                        candidate_text = key,
                        ref_text_list = self.ref_texts_keys,
                        candidate_text_model = None,
                        ref_text_model_list = self.ref_texts_chardiff_model,
                        model_params = self.textdiff_model_prms,
                        top_k = 5,
                    )
                    # self.logger.info('Similarity search for "' + str(key) + '", top distances ' + str(
                    #     top_distances) + ', object ' + str(object))
                    # raise Exception('asdf')
                    if top_keys:
                        if top_distances[0] <= difference_threshold:
                            key_similar = top_keys[0]
                            self.logger.info(
                                'Found via similarity search for "' + str(key) + '", a similar key "' + str(key_similar)
                                + '" distance ' + str(top_distances[0])
                            )
                            self.cache_dict[key_similar][self.KEY_REPEAT_COUNT] += 1
                            hit_similar = 1
                            return self.cache_dict[key_similar][self.KEY_VALUE]
                    return None
                else:
                    return None
        # except Exception as ex:
        #     self.logger.error('Unexpected error: ' + str(ex))
        #     raise Exception(ex)
        finally:
            self.update_cache_stats(hit_exact=hit_exact, hit_similar=hit_similar)
            self.__mutex_cache.release()

    def add_to_cache_threadsafe(
            self,
            object, # key
            result, # value
    ):
        if self.db_params is not None:
            obj_key = self.derive_key(object=object)
            rec = {
                self.KEY_ID: obj_key,
                self.KEY_VALUE: result,
                self.KEY_OBJECT_ORI: str(object),
                self.KEY_DATETIME: datetime.now().strftime(self.DATETIME_FORMAT),
            }
            self.cache_db.delete(
                tablename = self.db_params.db_table,
                match_phrase = {self.KEY_ID: obj_key},
            )
            self.cache_db.add(
                records = [rec],
                tablename = self.db_params.db_table,
            )
        else:
            return self.add_to_cache_dict_threadsafe(
                object = object,
                result = result,
            )

    def add_to_cache_dict_threadsafe(
            self,
            object, # key
            result, # value
    ):
        key = self.derive_key(object=object)

        try:
            self.__mutex_cache.acquire()

            is_cache_updated = False

            if len(self.cache_dict) >= self.cache_size:
                self.logger.info(
                    'Cache full at length ' + str(len(self.cache_dict)) + ' >= ' + str(self.cache_size)
                    + ', clear proportion ' + str(self.rm_prop_when_full) + '.'
                )
                if self.clear_cache_method == 'old':
                    # In this case, just clear oldest, which is FIFO, thus nothing to do, as already done below
                    pass
                else:
                    cache_tmp = self.cache_dict
                    self.cache_dict = {}
                    # remove all with no hits first
                    for k, v in cache_tmp.items():
                        if v[self.KEY_REPEAT_COUNT] > 0:
                            self.logger.info('Keep key "' + str(k) + '": ' + str(v))
                            self.cache_dict[k] = v
                        else:
                            self.logger.info('Discard key "' + str(k) + '": ' + str(v))
                    self.logger.info(
                        'Successfully cleanup up cache keeping only those with hits from length ' + str(len(cache_tmp))
                        + ', remaining items ' + str(len(self.cache_dict))
                    )

                # If still not hit the desired target, clear oldest ones
                count_desired = round(self.cache_size * (1 - self.rm_prop_when_full))
                if len(self.cache_dict) > count_desired:
                    # remove first added ones
                    count_thr = len(self.cache_dict) - count_desired
                    # Keep only latest added ones
                    self.cache_dict = {k:v for i,(k,v) in enumerate(self.cache_dict.items()) if i >= count_thr}
                    self.logger.warning(
                        'Further clear hit cache to new size ' + str(len(self.cache_dict)) + ', count thr ' + str(count_thr)
                        + ', max cache size ' + str(self.cache_size)
                    )
                is_cache_updated = True

            if is_cache_updated:
                self.__update_text_model()

            if key not in self.cache_dict.keys():
                self.cache_dict[key] = {
                    self.KEY_VALUE: result,
                    self.KEY_OBJECT_ORI: object,
                    self.KEY_REPEAT_COUNT: 0,
                    self.KEY_DATETIME: datetime.now()
                }
                self.ref_texts_keys.append(key)
                self.ref_texts_chardiff_model.append(
                    self.textdiff.get_text_model(
                        text = key,
                        model_params = self.textdiff_model_prms,
                    )
                )
            return key
        except Exception as ex:
            self.logger.error(
                'Error when adding key "' + str(key) + '" to cache: ' + str(ex)
            )
            return None
        finally:
            self.__mutex_cache.release()

    def __update_text_model(
            self,
    ):
        assert self.__mutex_cache.locked()

        self.ref_texts_keys = list(self.cache_dict.keys())
        self.ref_texts_chardiff_model = [
            self.textdiff.get_text_model(
                text = key,
                model_params = self.textdiff_model_prms,
            )
            for key in self.cache_dict.keys()
        ]
        self.logger.info('Successfully updated chardiff model')
        return

    def search_similar_object(
            self,
            text,
    ) -> tuple:
        try:
            self.__mutex_cache.acquire()
            return self.textdiff.text_difference(
                candidate_text = text,
                ref_text_list = self.ref_texts_keys,
                candidate_text_model = None,
                ref_text_model_list = self.ref_texts_chardiff_model,
                model_params = self.textdiff_model_prms,
                top_k = 5,
            )
        finally:
            self.__mutex_cache.release()


class TextSimilarityCacheUnitTest:
    def __init__(self, logger=None):
        self.logger = logger if logger is not None else logging.getLogger()
        return

    def test(self, additional_tests=()):
        self.test_cache_dict()
        if 'db' in additional_tests:
            self.test_cache_db()
        print('ALL TESTS PASSED')
        return

    def test_cache_db(self):
        data = [
            ('123456789012345678901234567890', '1'),
            ('1234567890123456789012345678901234567890', '2'),
        ]
        db_params = DbParams(
            identifier = 'test',
            db_type = os.environ["DB_TYPE"],
            db_host = os.environ["DB_HOST"],
            db_port = os.environ["DB_PORT"],
            db_username = os.environ["DB_USERNAME"],
            db_password = os.environ["DB_PASSWORD"],
            db_database = os.environ["DB_DATABASE"],
            db_table = "unittest_key_value_cache",
        )
        cache = TextSimilarityCache(
            db_params = db_params,
            cache_size = 4,
            clear_cache_method = 'unpopular',
            rm_prop_when_full = 0.5,
            logger = self.logger,
        )
        cache.clear_cache()
        for i, (obj, val) in enumerate(data):
            cache.add_to_cache_threadsafe(
                object = obj,
                result = val,
            )
            res = cache.get_from_cache_threadsafe(
                object = obj,
            )
            assert res == val, \
                'Test DB cache #' + str(i) + ' for object "' + str(obj) + '", got ' + str(res) + ' not ' + str(val)
        print('ALL CACHE DB TESTS PASSED')
        return

    def test_cache_dict(self):
        p = Profiling(logger=self.logger)
        data = [
            ('yo bro',           None,            ['yo bro']),
            ('dobriy dyen 1',    None,            ['yo bro', 'dobriy dyen 1']),
            ('kamusta 1',        None,            ['yo bro', 'dobriy dyen 1', 'kamusta 1']),
            ('dobriy dyen 1',    'dobriy dyen 1', ['yo bro', 'dobriy dyen 1', 'kamusta 1']), # exact hit
            # similarity search will return "dobriy dyen 1" above
            ('dobriy dyen 2',    'dobriy dyen 1', ['yo bro', 'dobriy dyen 1', 'kamusta 1', 'dobriy dyen 2']), # similar hit
            # At this point, we already reached unproven cache max size=4, and we clear those with no hits,
            # thus only ['dobriy dyen 1', 'kamusta 1'] remaining. 'kamusta 2' added after
            ('kamusta 2',        'kamusta 1',     ['dobriy dyen 1', 'kamusta 1', 'kamusta 2']), # similar hit
            ('kamusta 1',        'kamusta 1',     ['dobriy dyen 1', 'kamusta 1', 'kamusta 2']), # exact hit
            ('sami ludi 1',      None,            ['dobriy dyen 1', 'kamusta 1', 'kamusta 2', 'sami ludi 1']),
            # At this point, we already reached unproven cache max size=4, and we clear those with no hits,
            # thus only ['kamusta 1', 'sami ludi 1'] remaining. 'kamusta 2' added after
            ('sami ludi 1',      'sami ludi 1',   ['kamusta 1', 'sami ludi 1']), # exact hit
            ('sami ludi 2',      'sami ludi 1',   ['kamusta 1', 'sami ludi 1', 'sami ludi 2']), # similar hit
            ('sami ludi 3',      'sami ludi 1',   ['kamusta 1', 'sami ludi 1', 'sami ludi 2', 'sami ludi 3']), # similar hit
            # at this point cache will clear again keeping only those hit & latest ones added
            ('sami ludi 1',      'sami ludi 1',   ['kamusta 1', 'sami ludi 1']), # exact hit
        ]
        cache = TextSimilarityCache(
            db_params = None,
            cache_size = 4,
            clear_cache_method = 'unpopular',
            rm_prop_when_full = 0.5,
            logger = self.logger,
        )
        cache.clear_cache()

        count_hit_exact = 0
        count_hit_similar = 0

        for i, tp in enumerate(data):
            p.start_time_profiling(id='ut')
            # print('At line #' + str(i+1) + '. ' + str(data[i]) + ', cache state ' + str(cache.cache_db))

            txt, expected_res_from_cache, expected_proven_cache_keys = tp
            res = cache.get_from_cache_threadsafe(
                object = txt,
                difference_threshold = 0.2,
            )
            cache.add_to_cache_threadsafe(
                object = txt,
                result = txt,
            )
            if res is not None:
                if res == txt:
                    print('Exact hit with "' + str(res) + '" == "' + str(txt) + '"')
                    count_hit_exact += 1
                else:
                    print('Similar hit with "' + str(res) + '" != "' + str(txt) + '"')
                    count_hit_similar += 1
            # print('added key:', added_key, ', text', txt, ', result', res, ', expected result', expected_result)
            # print('#' + str(i) + '. Proven cache:\n' + str(cache.cache_proven) + ', unproven cache:\n' + str(cache.cache_unproven))
            assert res == expected_res_from_cache, \
                '#' + str(i+1) + '. Get result "' + str(res) + '" not expected result "' \
                + str(expected_res_from_cache) + '" for text "' + str(txt) + '", test data ' + str(tp)
            cur_keys = cache.get_cache_keys()
            diff_set = list(set(cur_keys).difference(set(expected_proven_cache_keys)))
            assert diff_set == [], \
                '#' + str(i+1) + '. Cache keys ' + str(cur_keys) \
                + '" not expected ' + str(expected_proven_cache_keys) + '" for text "' + str(txt) \
                + '", test data ' + str(tp)

            # print('(AFTER ADD) At line #' + str(i+1) + '. cache state ' + str(cache.cache_db))
            top_keys, top_dists = cache.search_similar_object(text=txt)
            self.logger.info('Sim search "' + str(txt) + '": ' + str(list(zip(top_keys, top_dists))))

            p.record_time_profiling(id='ut', msg=data[i], logmsg=True)

        self.logger.info(cache.cache_stats)
        assert cache.cache_stats['total'] == len(data), \
            'Data length ' + str(len(data)) + ' but cache total ' + str(cache.cache_stats['total'])
        assert cache.cache_stats['hit_exact'] == count_hit_exact == 4, \
            'Exact hits ' + str(cache.cache_stats['hit_exact'])
        assert cache.cache_stats['hit_similar'] == count_hit_similar == 4, \
            'Similar hits ' + str(cache.cache_stats['hit_similar'])

        # Old test
        # Cache max 4, and remove 1 when full everytime
        cache = TextSimilarityCache(
            cache_size = 4,
            clear_cache_method = 'old',
            rm_prop_when_full = 0.3,
            logger = self.logger,
        )
        for i in range(10):
            cache.add_to_cache_threadsafe(object=i, result=10 * i)
            keys = cache.get_cache_keys()
            self.logger.info('***** ' + str(i) + str(cache.cache_dict))
            keys_expected = [str(v) for v in list(range(i + 1)) if i - v < 4]
            assert keys == keys_expected, 'Keys ' + str(keys) + ' not expected ' + str(keys_expected)

        self.logger.info('ALL CACHE DICT TESTS PASSED')
        return


if __name__ == '__main__':
    lgr = Logging.get_default_logger(log_level=logging.INFO, propagate=False)
    er = Env(logger=lgr)
    Env.set_env_vars_from_file(env_filepath=er.REPO_DIR + '/.env.fitxf.math.ut')

    TextSimilarityCacheUnitTest(logger=lgr).test(additional_tests=[])

    exit(0)



import logging
import torch
import threading
import numpy as np
from sklearn.linear_model import LinearRegression
from fitxf import FitUtils
from fitxf.utils import Logging


class Regression(torch.nn.Module):

    def __init__(
            self,
            # by default is a cubic polynomial
            polynomial_order: int = 3,
            learning_rate: float = 0.001,
            logger: Logging = None,
    ):
        super().__init__()
        self.polynomial_order = polynomial_order
        self.learning_rate = learning_rate
        self.logger = logger if logger is not None else logging.getLogger()

        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        self.fit_utils = FitUtils(logger=self.logger)
        self.init_network_arc()
        return

    def fit_linear_regression(self, X: np.ndarray, y: np.ndarray):
        rg = LinearRegression().fit(X, y)
        coef = rg.coef_
        intercept = rg.intercept_
        return coef, intercept

    def predict_linear_regression(self, X: np.ndarray, coef: np.ndarray, intercept: np.ndarray):
        return intercept + coef * X

    def init_network_arc(self):
        self.layer_1 = torch.nn.Linear(in_features=1+self.polynomial_order, out_features=1)

        # Random initialization of model parameters if not loading from a previous state
        for p in self.parameters():
            if p.dim() > 1:
                torch.nn.init.xavier_uniform_(p)

        self.loss_func = torch.nn.MSELoss()
        self.optimizer = torch.optim.Adam(
            self.parameters(),
            lr = self.learning_rate,
            betas = (0.9, 0.98),
            eps = 1e-9
        )
        return

    def split_data_train(
            self,
            X: torch.Tensor,
            y: torch.Tensor,
            # if None, means we don't convert to onehot (possibly caller already done that, or not required)
            num_categories: int = None,
            # the smaller the batch size, the smaller the losses will be during training
            batch_size: int = 32,
            eval_percent: float = 0.2,
            include_attn_mask: bool = False,
    ):
        dl_trn, dl_val, n_cutoff_train = self.fit_utils.create_dataloader(
            X = X,
            y = y,
            y_num_classes = num_categories,
            batch_size = batch_size,
            eval_percent = eval_percent,
            include_attn_mask = include_attn_mask,
        )
        self.logger.info(
            'Train length = ' + str(len(dl_trn)) + ', eval length = ' + str(len(dl_val))
            + ', cutoff train = ' +  str(n_cutoff_train)
        )
        self.logger.info('Data loader train ' + str(dl_trn))
        # raise Exception('asdf')
        return dl_trn, dl_val, n_cutoff_train

    def fit_custom_func(
            self,
            X: torch.Tensor,
            y: torch.Tensor,
            # with interfaces forward(), parameters() & derivative()
            custom_func,
            loss_func,
            batch_size: int = 32,
            eval_percent: float = 0.2,
            epochs: int = 100,
    ):
        dl_trn, dl_val, n_cutoff_train = self.split_data_train(
            X = X,
            y = y,
            num_categories = None,
            batch_size = batch_size,
            eval_percent = eval_percent,
        )
        for i in range(epochs):
            y_round = custom_func.forward(X)
            loss_val = np.sum((y_round - y) ** 2)


    # func: function callback f(X) -> y
    def fit(
            self,
            X: torch.Tensor,
            y: torch.Tensor,
            is_categorical: bool = False,
            # if None, means we don't convert to onehot (possibly caller already done that, or not required)
            num_categories: int = None,
            # the smaller the batch size, the smaller the losses will be during training
            batch_size: int = 32,
            eval_percent: float = 0.2,
            epochs: int = 100,
            # important to prevent over-fitting
            regularization_type = None,
    ):
        dl_trn, dl_val, n_cutoff_train = self.split_data_train(
            X = X,
            y = y,
            num_categories = num_categories,
            batch_size = batch_size,
            eval_percent = eval_percent,
        )

        losses = self.fit_utils.torch_train(
            model = self,
            train_dataloader = dl_trn,
            loss_func = self.loss_func,
            optimizer = self.optimizer,
            regularization_type = regularization_type,
            epochs = epochs,
        )
        self.logger.info('Train losses: ' + str(losses))
        # Important! Set back to eval mode, so subsequent forwards won't affect any weights or gradients
        self.eval()

        if eval_percent > 0:
            eval_accuracy = self.evaluate_accuracy(
                X_eval = X[n_cutoff_train:],
                y_eval = y[n_cutoff_train:],
                is_categorical = is_categorical,
                num_categories = num_categories,
            )
        else:
            eval_accuracy = None

        return {
            'eval_accuracy': eval_accuracy,
            'losses': losses,
            'dataloader_train': dl_trn,
            'dataloader_eval': dl_val,
            'index_cutoff_train': n_cutoff_train,
        }

    def forward(
            self,
            X: torch.Tensor,
    ):
        return self.layer_1(X)


class RegressionUnitTest:
    def __init__(self, logger: Logging = None):
        self.logger = logger if logger is not None else logging.getLogger()
        self.regressor = Regression(logger=self.logger)
        self.custom_func = self.CustomFunctionExample(logger=self.logger)
        return

    class CustomLossMseExample:
        def __init__(self, logger: Logging = None):
            return

        def forward(self, y_eval: torch.Tensor, y: torch.Tensor):
            return np.sum((y_eval - y) ** 2)

    class CustomFunctionExample:
        def __init__(self, logger: Logging = None):
            self.logger = logger if logger is not None else logging.getLogger()
            self.params = torch.Tensor([2, 0.1])
            self.params_delta = torch.zeros(size=self.params.size())
            assert self.params.size() == self.params_delta.size()
            self.params_shape = self.params.size()
            return

        def get_params(self):
            return self.params

        def update_params(self, new_params: torch.Tensor):
            assert new_params.size() == self.params_shape
            self.params = new_params

        def forward(self, X: torch.Tensor, params: torch.Tensor = None):
            prms = self.params if params is None else params
            assert prms.size() == self.params_shape
            A, k = prms
            return A * torch.exp(-k * X)

        def derivative(self, X: torch.Tensor, delta: float = 0.001):
            derv = [None]*len(self.params )
            for i, _ in enumerate(self.params):
                prms_i_delta = self.params.clone().detach()
                prms_i_delta[i] = prms_i_delta[i] + delta
                # TODO need to normalize?
                derv_i = self.forward(X=X, params=prms_i_delta) - self.forward(X=X, params=self.params)
                self.logger.info(
                    'Derivative for i=' + str(i) + ', ' + str(derv_i) + ', params delta ' + str(prms_i_delta)
                    + ', params ' + str(self.params)
                )
                derv[i] = derv_i.tolist()
            tsr_derv = torch.Tensor(derv)
            self.logger.info('Final derivative ' + str(tsr_derv))
            return tsr_derv

    def test_linear_regression(self):
        X = np.array([[80], [65], [50], [30], [10]])
        y = np.array([6, 5, 4, 3, 2])
        cf, ic = self.regressor.fit_linear_regression(X=X, y=y)
        self.logger.info('Type ' + str(type(cf)) + ', ' + str(type(ic)))
        exp_cf, exp_ic = np.array([0.05681818]), 1.3295454545454541
        assert np.sum((cf - exp_cf)**2) < 0.0000000001, 'Fit coefficient ' + str(cf) + ', not expected ' + str(exp_cf)
        assert ic == exp_ic, 'Fit intercept ' + str(ic) + ' not expected ' + str(exp_ic)

        self.logger.info('Coef ' + str(cf) + ', intercept ' + str(ic))
        y_ = self.regressor.predict_linear_regression(X=X, coef=cf, intercept=ic)
        exp_y_pred = [[5.875     ], [5.02272727], [4.17045455], [3.03409091], [1.89772727]]
        assert np.sum((y_ - exp_y_pred)**2) < 0.0000000001, \
            'Predicted values ' + str(y_) + ', not expected ' + str(exp_y_pred)
        self.logger.info('Linear Regression Tests Passed')
        return

    def test(self):
        self.test_linear_regression()

        X = np.array([[80], [65], [50], [30], [10]])
        f = self.CustomFunctionExample(logger=self.logger)
        self.logger.info(f.forward(X))
        self.logger.info(f.derivative(X))

        self.logger.info('All Tests Passed')


if __name__ == '__main__':
    lgr = Logging.get_default_logger(log_level=logging.INFO, propagate=False)
    RegressionUnitTest(logger=lgr).test()
    exit(0)

    order = 1
    r_poly = Regression(
        polynomial_order = order,
        learning_rate = 0.01,
        logger = lgr,
    )
    X_poly = torch.Tensor([
        [x[0]**i for i in range(order+1)] for x in X
    ])
    y_poly = torch.from_numpy(y).to(torch.float)
    lgr.info(X_poly)
    r_poly.fit(
        X = X_poly,
        y = y_poly,
        eval_percent = 0.,
        batch_size = len(X_poly),
        epochs = 100,
        # regularization_type = "L1",
    )
    lgr.info('Model parameters ' + str(r_poly.state_dict()))
    lgr.info('y predict ' + str(r_poly(X_poly)))

    exit(0)



import logging
import torch
from fitxf.utils import Logging


class FunctionInterface:

    def __init__(self, params: torch.Tensor = None, logger: Logging = None):
        self.logger = logger if logger is not None else logging.getLogger()
        self.params = params
        return

    def forward(self, X: torch.Tensor, params: torch.Tensor = None) -> torch.Tensor:
        raise Exception('Must be implemented by child class!!')

    def derivative_of_params(self, X: torch.Tensor, delta: float = 0.001) -> torch.Tensor:
        if self.params is None:
            return torch.DoubleTensor([0.])

        derv = [None] * len(self.params)
        for i, _ in enumerate(self.params):
            prms_i_delta = self.params.clone().detach()
            prms_i_delta[i] = prms_i_delta[i] + delta
            # TODO need to normalize?
            f_X_d = self.forward(X=X, params=prms_i_delta)
            f_X_0 = self.forward(X=X, params=self.params)
            self.logger.debug('f(X, params+delta) = ' + str(f_X_d) + ' for X ' + str(X))
            self.logger.debug('f(X, params) = ' + str(f_X_0) + ' for X ' + str(X))
            derv_i = (f_X_d - f_X_0)
            self.logger.debug(
                'Derivative for i=' + str(i) + ', ' + str(derv_i) + ', params delta ' + str(prms_i_delta)
                + ', params ' + str(self.params)
            )
            derv[i] = derv_i.tolist()
        tsr_derv = torch.DoubleTensor(derv)
        tsr_derv = torch.transpose(tsr_derv, 0, 1)
        self.logger.info('Final derivative ' + str(tsr_derv))
        return tsr_derv


if __name__ == '__main__':
    lgr = Logging.get_default_logger(log_level=logging.DEBUG, propagate=False)
    class F(FunctionInterface):
        def __init__(self, params: torch.Tensor = None, logger: Logging = None):
            super().__init__(params=params, logger=logger)

        def forward(self, X: torch.Tensor, params: torch.Tensor = None) -> torch.Tensor:
            prms = self.params if params is None else params
            return prms[0] + prms[1]*X + prms[2]*(X**2)

    params = torch.Tensor([2.0, 1.0, 0.5])
    X_test = torch.Tensor([[-1], [0]])
    f = F(params=params, logger=lgr)
    lgr.info(f.forward(X=X_test))
    lgr.info(f.derivative_of_params(X=X_test, delta=0.1))
    exit(0)



import logging
import numpy as np
from fitxf.utils import Logging


# https://en.wikipedia.org/wiki/%CE%9C-law_algorithm
class Mulaw:

    BIN_INTERVALS = (
        # high/low, low/high, num sub intervals, code
        ( 8158, 4063, 16, 0x80),
        ( 4062, 2015, 16, 0x90),
        ( 2014,  991, 16, 0xA0),
        (  990,  479, 16, 0xB0),
        (  478,  223, 16, 0xC0),
        (  222,   95, 16, 0xD0),
        (   94,   31, 16, 0xE0),
        (   30,    1, 15, 0xF0),
        (    0,    0,  1, 0xFF),
        (   -1,   -1,  1, 0x7F),
        (  -31,   -2, 15, 0x70),
        (  -95,  -32, 16, 0x60),
        ( -223,  -96, 16, 0x50),
        ( -479, -224, 16, 0x40),
        ( -991, -480, 16, 0x30),
        (-2015, -992, 16, 0x20),
        (-4063,-2016, 16, 0x10),
        (-8159,-4064, 16, 0x00),
    )
    MAX_VAL = 8158

    def __init__(self, logger: Logging = None):
        self.logger = logger if logger is not None else logging.getLogger()
        self.create_bins()
        return

    def create_bins(self):
        self.edge_bins = np.array([v[0] for v in self.BIN_INTERVALS])[::-1]
        self.code_bins = np.array([v[3] for v in self.BIN_INTERVALS])[::-1]
        self.logger.info('Edges\n' + str(self.edge_bins))
        self.logger.info('Codes\n' + str(self.code_bins))
        return

    def u_law_enc(self, x, mu = 255):
        assert np.max(np.abs(x)) <= 1
        sgn = -1 * (x < 0) + 1 * (x >= 0)
        y = sgn * np.log(1 + mu * np.abs(x)) / np.log(1 + mu)
        return y

    def u_law_dec(self, y, mu = 255):
        sgn = -1 * (y < 0) + 1 * (y >= 0)
        x = sgn * ( (1 + mu) ** np.abs(y) - 1) / mu
        return x


if __name__ == '__main__':
    lgr = Logging.get_default_logger(log_level=logging.DEBUG, propagate=False)
    ml = Mulaw(logger=lgr)
    x = 2 * ( (np.arange(101) / 100) - 0.5)
    lgr.info(x)
    x_enc = ml.u_law_enc(x=x)
    x_dec = ml.u_law_dec(y=x_enc)
    tmp = np.array([
        x.tolist(), x_enc.tolist(), x_dec.tolist()
    ]).transpose()
    lgr.info(tmp)
    exit(0)
